package vpc

import (
	"bytes"
	"crypto/md5"
	"fmt"
	"net"

	"github.com/dropbox/godropbox/container/set"
	"github.com/pritunl/mongo-go-driver/bson"
	"github.com/pritunl/mongo-go-driver/bson/primitive"
	"github.com/pritunl/mongo-go-driver/mongo/options"
	"github.com/pritunl/pritunl-cloud/database"
	"github.com/pritunl/pritunl-cloud/utils"
)

func GetIp6(vpcId primitive.ObjectID, addr net.IP) net.IP {
	netHash := md5.New()
	netHash.Write(vpcId[:])
	netHashSum := fmt.Sprintf("%x", netHash.Sum(nil))[:12]

	macHash := md5.New()
	macHash.Write(addr)
	macHashSum := fmt.Sprintf("%x", macHash.Sum(nil))[:16]

	// fd97:2222:2222:2222:3333:3333:4444:4444
	// 2 = vpc
	// 3 = subnet
	// 4 = instance+gateway

	ip := fmt.Sprintf("fd97%s%s", netHashSum, macHashSum)
	ipBuf := bytes.Buffer{}

	for i, run := range ip {
		if i%4 == 0 && i != 0 && i != len(ip)-1 {
			ipBuf.WriteRune(':')
		}
		ipBuf.WriteRune(run)
	}

	return net.ParseIP(ipBuf.String())
}

func GetGatewayIp6(vpcId primitive.ObjectID, addr net.IP) net.IP {
	netHash := md5.New()
	netHash.Write(vpcId[:])
	netHashSum := fmt.Sprintf("%x", netHash.Sum(nil))[:12]

	macHash := md5.New()
	macHash.Write(addr)
	macHash.Write(addr)
	macHashSum := fmt.Sprintf("%x", macHash.Sum(nil))[:16]

	// fd97:2222:2222:2222:3333:3333:4444:4444
	// 2 = vpc
	// 3 = subnet
	// 4 = instance+gateway

	ip := fmt.Sprintf("fd97%s%s", netHashSum, macHashSum)
	ipBuf := bytes.Buffer{}

	for i, run := range ip {
		if i%4 == 0 && i != 0 && i != len(ip)-1 {
			ipBuf.WriteRune(':')
		}
		ipBuf.WriteRune(run)
	}

	return net.ParseIP(ipBuf.String())
}

func GetLinkIp6(vpcId primitive.ObjectID, addr net.IP) net.IP {
	netHash := md5.New()
	netHash.Write(vpcId[:])
	netHashSum := fmt.Sprintf("%x", netHash.Sum(nil))[:12]

	macHash := md5.New()
	macHash.Write(addr)
	macHashSum := fmt.Sprintf("%x", macHash.Sum(nil))[:16]

	// fe80:2222:2222:2222:3333:3333:4444:4444
	// 2 = vpc
	// 3 = subnet
	// 4 = instance+gateway

	ip := fmt.Sprintf("fe80%s%s", netHashSum, macHashSum)
	ipBuf := bytes.Buffer{}

	for i, run := range ip {
		if i%4 == 0 && i != 0 && i != len(ip)-1 {
			ipBuf.WriteRune(':')
		}
		ipBuf.WriteRune(run)
	}

	return net.ParseIP(ipBuf.String())
}

func GetGatewayLinkIp6(vpcId primitive.ObjectID, addr net.IP) net.IP {
	netHash := md5.New()
	netHash.Write(vpcId[:])
	netHashSum := fmt.Sprintf("%x", netHash.Sum(nil))[:12]

	macHash := md5.New()
	macHash.Write(addr)
	macHash.Write(addr)
	macHashSum := fmt.Sprintf("%x", macHash.Sum(nil))[:16]

	// fe80:2222:2222:2222:3333:3333:4444:4444
	// 2 = vpc
	// 3 = subnet
	// 4 = instance+gateway

	ip := fmt.Sprintf("fe80%s%s", netHashSum, macHashSum)
	ipBuf := bytes.Buffer{}

	for i, run := range ip {
		if i%4 == 0 && i != 0 && i != len(ip)-1 {
			ipBuf.WriteRune(':')
		}
		ipBuf.WriteRune(run)
	}

	return net.ParseIP(ipBuf.String())
}

func Get(db *database.Database, vcId primitive.ObjectID) (
	vc *Vpc, err error) {

	coll := db.Vpcs()
	vc = &Vpc{}

	err = coll.FindOneId(vcId, vc)
	if err != nil {
		return
	}

	return
}

func GetOrg(db *database.Database, orgId, vcId primitive.ObjectID) (
	vc *Vpc, err error) {

	coll := db.Vpcs()
	vc = &Vpc{}

	err = coll.FindOne(db, &bson.M{
		"_id":          vcId,
		"organization": orgId,
	}).Decode(vc)
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func ExistsOrg(db *database.Database, orgId, vcId primitive.ObjectID) (
	exists bool, err error) {

	coll := db.Vpcs()
	n, err := coll.CountDocuments(
		db,
		&bson.M{
			"_id":          vcId,
			"organization": orgId,
		},
	)
	if err != nil {
		return
	}

	if n > 0 {
		exists = true
	}

	return
}

func GetAll(db *database.Database, query *bson.M) (
	vcs []*Vpc, err error) {

	coll := db.Vpcs()
	vcs = []*Vpc{}

	cursor, err := coll.Find(
		db,
		query,
	)
	if err != nil {
		err = database.ParseError(err)
		return
	}
	defer cursor.Close(db)

	for cursor.Next(db) {
		vc := &Vpc{}
		err = cursor.Decode(vc)
		if err != nil {
			err = database.ParseError(err)
			return
		}

		vcs = append(vcs, vc)
	}

	err = cursor.Err()
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func GetAllNames(db *database.Database, query *bson.M) (
	vpcs []*Vpc, err error) {

	coll := db.Vpcs()
	vpcs = []*Vpc{}

	cursor, err := coll.Find(
		db,
		query,
		&options.FindOptions{
			Sort: &bson.D{
				{"name", 1},
			},
			Projection: &bson.D{
				{"name", 1},
				{"organization", 1},
				{"datacenter", 1},
				{"type", 1},
				{"subnets", 1},
			},
		},
	)
	if err != nil {
		err = database.ParseError(err)
		return
	}
	defer cursor.Close(db)

	for cursor.Next(db) {
		vc := &Vpc{}
		err = cursor.Decode(vc)
		if err != nil {
			err = database.ParseError(err)
			return
		}

		vpcs = append(vpcs, vc)
	}

	err = cursor.Err()
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func GetAllPaged(db *database.Database, query *bson.M,
	page, pageCount int64) (vcs []*Vpc, count int64, err error) {

	coll := db.Vpcs()
	vcs = []*Vpc{}

	if len(*query) == 0 {
		count, err = coll.EstimatedDocumentCount(db)
		if err != nil {
			err = database.ParseError(err)
			return
		}
	} else {
		count, err = coll.CountDocuments(db, query)
		if err != nil {
			err = database.ParseError(err)
			return
		}
	}

	maxPage := count / pageCount
	if count == pageCount {
		maxPage = 0
	}
	page = utils.Min64(page, maxPage)
	skip := utils.Min64(page*pageCount, count)

	cursor, err := coll.Find(
		db,
		query,
		&options.FindOptions{
			Sort: &bson.D{
				{"name", 1},
			},
			Skip:  &skip,
			Limit: &pageCount,
		},
	)
	if err != nil {
		err = database.ParseError(err)
		return
	}
	defer cursor.Close(db)

	for cursor.Next(db) {
		vc := &Vpc{}
		err = cursor.Decode(vc)
		if err != nil {
			err = database.ParseError(err)
			return
		}

		vcs = append(vcs, vc)
	}

	err = cursor.Err()
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func GetIds(db *database.Database, ids []primitive.ObjectID) (
	vcs []*Vpc, err error) {

	coll := db.Vpcs()
	vcs = []*Vpc{}

	cursor, err := coll.Find(
		db,
		&bson.M{
			"_id": &bson.M{
				"$in": ids,
			},
		},
	)
	if err != nil {
		err = database.ParseError(err)
		return
	}
	defer cursor.Close(db)

	for cursor.Next(db) {
		vc := &Vpc{}
		err = cursor.Decode(vc)
		if err != nil {
			err = database.ParseError(err)
			return
		}

		vcs = append(vcs, vc)
	}

	err = cursor.Err()
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func GetDatacenter(db *database.Database, dcId primitive.ObjectID) (
	vcs []*Vpc, err error) {

	coll := db.Vpcs()
	vcs = []*Vpc{}

	cursor, err := coll.Find(
		db,
		&bson.M{
			"datacenter": dcId,
		},
	)
	if err != nil {
		err = database.ParseError(err)
		return
	}
	defer cursor.Close(db)

	for cursor.Next(db) {
		vc := &Vpc{}
		err = cursor.Decode(vc)
		if err != nil {
			err = database.ParseError(err)
			return
		}

		vcs = append(vcs, vc)
	}

	err = cursor.Err()
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func DistinctIds(db *database.Database, matchIds []primitive.ObjectID) (
	idsSet set.Set, err error) {

	coll := db.Images()
	idsSet = set.NewSet()

	idsInf, err := coll.Distinct(
		db,
		"_id",
		&bson.M{
			"_id": &bson.M{
				"$in": matchIds,
			},
		},
	)
	if err != nil {
		err = database.ParseError(err)
		return
	}

	for _, idInf := range idsInf {
		if id, ok := idInf.(primitive.ObjectID); ok {
			idsSet.Add(id)
		}
	}

	return
}

func Remove(db *database.Database, vcId primitive.ObjectID) (err error) {
	coll := db.VpcsIp()

	_, err = coll.DeleteMany(db, &bson.M{
		"vpc": vcId,
	})
	if err != nil {
		err = database.ParseError(err)
		return
	}

	coll = db.Vpcs()

	_, err = coll.DeleteOne(db, &bson.M{
		"_id": vcId,
	})
	if err != nil {
		err = database.ParseError(err)
		switch err.(type) {
		case *database.NotFoundError:
			err = nil
		default:
			return
		}
	}

	return
}

func RemoveOrg(db *database.Database, orgId, vcId primitive.ObjectID) (
	err error) {

	coll := db.VpcsIp()

	_, err = coll.DeleteMany(db, &bson.M{
		"vpc": vcId,
	})
	if err != nil {
		err = database.ParseError(err)
		return
	}

	coll = db.Vpcs()

	_, err = coll.DeleteOne(db, &bson.M{
		"organization": orgId,
		"_id":          vcId,
	})
	if err != nil {
		err = database.ParseError(err)
		switch err.(type) {
		case *database.NotFoundError:
			err = nil
		default:
			return
		}
	}

	return
}

func RemoveMulti(db *database.Database, vcIds []primitive.ObjectID) (err error) {
	coll := db.VpcsIp()

	_, err = coll.DeleteMany(db, &bson.M{
		"vpc": &bson.M{
			"$in": vcIds,
		},
	})
	if err != nil {
		err = database.ParseError(err)
		return
	}

	coll = db.Vpcs()

	_, err = coll.DeleteMany(db, &bson.M{
		"_id": &bson.M{
			"$in": vcIds,
		},
	})
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func GetIpsMapped(db *database.Database, ids []primitive.ObjectID) (
	vpcsMap map[primitive.ObjectID][]*VpcIp, err error) {

	coll := db.VpcsIp()
	vpcsMap = map[primitive.ObjectID][]*VpcIp{}

	cursor, err := coll.Find(
		db,
		&bson.M{
			"vpc": &bson.M{
				"$in": ids,
			},
		},
	)
	if err != nil {
		err = database.ParseError(err)
		return
	}
	defer cursor.Close(db)

	for cursor.Next(db) {
		vc := &VpcIp{}
		err = cursor.Decode(vc)
		if err != nil {
			err = database.ParseError(err)
			return
		}

		vpcsMap[vc.Vpc] = append(vpcsMap[vc.Vpc], vc)
	}

	err = cursor.Err()
	if err != nil {
		err = database.ParseError(err)
		return
	}

	return
}

func RemoveInstanceIps(db *database.Database, instId primitive.ObjectID) (
	err error) {

	coll := db.VpcsIp()

	_, err = coll.UpdateMany(db, &bson.M{
		"instance": instId,
	}, &bson.M{
		"$set": &bson.M{
			"instance": nil,
		},
	})
	if err != nil {
		err = database.ParseError(err)
		switch err.(type) {
		case *database.NotFoundError:
			err = nil
		default:
			return
		}
	}

	return
}

func RemoveInstanceIp(db *database.Database, instId,
	vpcId primitive.ObjectID) (err error) {

	coll := db.VpcsIp()

	_, err = coll.UpdateOne(
		db,
		&bson.M{
			"vpc":      vpcId,
			"instance": instId,
		},
		&bson.M{
			"$set": &bson.M{
				"instance": nil,
			},
		},
	)
	if err != nil {
		err = database.ParseError(err)
		switch err.(type) {
		case *database.NotFoundError:
			err = nil
		default:
			return
		}
	}

	return
}
