/// <reference path="../References.d.ts"/>
import * as SuperAgent from 'superagent';
import Dispatcher from '../dispatcher/Dispatcher';
import EventDispatcher from '../dispatcher/EventDispatcher';
import * as Alert from '../Alert';
import * as Csrf from '../Csrf';
import Loader from '../Loader';
import * as GlobalTypes from '../types/GlobalTypes';
import * as OrganizationTypes from '../types/OrganizationTypes';
import * as MiscUtils from '../utils/MiscUtils';

let syncId: string;

export function sync(): Promise<void> {
	let curSyncId = MiscUtils.uuid();
	syncId = curSyncId;

	let loader = new Loader().loading();

	return new Promise<void>((resolve, reject): void => {
		SuperAgent
			.get('/organization')
			.set('Accept', 'application/json')
			.set('Csrf-Token', Csrf.token)
			.end((err: any, res: SuperAgent.Response): void => {
				loader.done();

				if (res && res.status === 401) {
					window.location.href = '/login';
					resolve();
					return;
				}

				if (curSyncId !== syncId) {
					resolve();
					return;
				}

				if (err) {
					Alert.errorRes(res, 'Failed to load organizations');
					reject(err);
					return;
				}

				Dispatcher.dispatch({
					type: OrganizationTypes.SYNC,
					data: {
						organizations: res.body,
					},
				});

				resolve();
			});
	});
}

export function commit(org: OrganizationTypes.Organization): Promise<void> {
	let loader = new Loader().loading();

	return new Promise<void>((resolve, reject): void => {
		SuperAgent
			.put('/organization/' + org.id)
			.send(org)
			.set('Accept', 'application/json')
			.set('Csrf-Token', Csrf.token)
			.end((err: any, res: SuperAgent.Response): void => {
				loader.done();

				if (res && res.status === 401) {
					window.location.href = '/login';
					resolve();
					return;
				}

				if (err) {
					Alert.errorRes(res, 'Failed to save organization');
					reject(err);
					return;
				}

				resolve();
			});
	});
}

export function create(org: OrganizationTypes.Organization): Promise<void> {
	let loader = new Loader().loading();

	return new Promise<void>((resolve, reject): void => {
		SuperAgent
			.post('/organization')
			.send(org)
			.set('Accept', 'application/json')
			.set('Csrf-Token', Csrf.token)
			.end((err: any, res: SuperAgent.Response): void => {
				loader.done();

				if (res && res.status === 401) {
					window.location.href = '/login';
					resolve();
					return;
				}

				if (err) {
					Alert.errorRes(res, 'Failed to create organization');
					reject(err);
					return;
				}

				resolve();
			});
	});
}

export function remove(orgId: string): Promise<void> {
	let loader = new Loader().loading();

	return new Promise<void>((resolve, reject): void => {
		SuperAgent
			.delete('/organization/' + orgId)
			.set('Accept', 'application/json')
			.set('Csrf-Token', Csrf.token)
			.end((err: any, res: SuperAgent.Response): void => {
				loader.done();

				if (res && res.status === 401) {
					window.location.href = '/login';
					resolve();
					return;
				}

				if (err) {
					Alert.errorRes(res, 'Failed to delete organizations');
					reject(err);
					return;
				}

				resolve();
			});
	});
}

export function setCurrent(current: string): void {
	Dispatcher.dispatch({
		type: GlobalTypes.RESET,
		data: {
			current: current,
		},
	});
	Dispatcher.dispatch({
		type: GlobalTypes.RELOAD,
		data: {
			current: current,
		},
	});
}

EventDispatcher.register((action: OrganizationTypes.OrganizationDispatch) => {
	switch (action.type) {
		case OrganizationTypes.CHANGE:
			sync();
			break;
	}
});
